/* This include file to the POV raytracer defines a MESH of REVOLUTION object

   -------------------------------------------------------------------------------------------        

   #macro mesh_of_revolution(points, angular_range, division, smoothing, closed)
                                                              
        The macro returns a MESH derived from the rotation of an array of uv-vectors 'points'
        for the given 'angular_range' degrees around the y-axis. It is similar to the original 
        SOR (surface of rotation) object of POV_Ray except for the following:
           - renders error free in POV_Ray 3.1g
           - allows the intersection of the 2D-curve defined by 'points' with itself.
        The u component of the 'points' vectors gives the distance from the y-axis while
        the v component gives the length along this axis.
        The number given by the parameter 'division' gives the number of angular steps for 
        the given 'angular_range'.  
        The parameter 'smoothing' allows surface smoothing by the use of 'smooth_triangle' 
        (smoothing=1) instead of 'triangle' (smoothing=0). The normal vector for each point 
        on the definition curve is derived from its two neighbours. 
        In any case, with or without smoothing, the first AND last point in the array
        points is NOT on the surface of the mesh. It is only used to derive the surface
        normal of the second and second last point - thus the minimum size of the array
        'points' is 4 (four) uv-vectors
        The parameter 'closed' closes the "top" and "bottom" end of the figure with a plane.
        As this is derived from the second and second-last point in the list of points, it is
        not necessarily the real bottom and top: Remember that the curve is allowed to 
        intersect itself!

   #macro union_of_revolution(points, angular_range, division, smoothing, closed)
   
        mostly the same macro but it returns a union of the face objects instead of a triangular mesh.
 
See the files mordemo.pov for some examples.
           
Disclaimer:    
    
                  This macro is free for use together with the POVray raytracer.
          It should be distributed with the files geometry.txt and demo_01.pov and demo_02.pov.

                                                     Uwe Zimmermann, Stockholm, Sweden 1999.
                                                                  
   -------------------------------------------------------------------------------------------     */



#macro mor_do_revolution(points, angular_range, division, smoothing, closed)
  #local n_points=dimension_size(points, 1);
  #if (n_points<4) 
    #error "too few points in mesh_of-revolution!\n"
  #else                                           
    #local normals=array[dimension_size(points, 1)-1]
    #local i=1;
    #while (i<n_points-1)
      #local dist_v1 = <points[i].u-points[i-1].u,points[i].v-points[i-1].v>;                                           
      #local dist_v2 = <points[i+1].u-points[i].u,points[i+1].v-points[i].v>;                                           
      #local dist_1  = vlength(dist_v1);
      #local dist_2  = vlength(dist_v2);
      #local normals[i] = <dist_2*dist_v1.v + dist_1*dist_v2.v,dist_2*dist_v1.u + dist_1*dist_v2.u>/(dist_1+dist_2);
      #local i=i+1;
    #end  
    #local i=1;
    #while (i<n_points-2)
      #local j=0;                       
      #while (j<=division)                      
        #local cos_phi=cos(j*2*pi*angular_range/360/division); 
        #local sin_phi=sin(j*2*pi*angular_range/360/division); 
        #local cos_phi2=cos((j+1)*2*pi*angular_range/360/division); 
        #local sin_phi2=sin((j+1)*2*pi*angular_range/360/division); 
        #if ((i=1)&(closed=1))
          triangle { 
                     <0, points[i].v, 0>
                     <cos_phi*points[i].u, points[i].v, sin_phi*points[i].u>
                     <cos_phi2*points[i].u, points[i].v, sin_phi2*points[i].u>
                   }
          triangle { 
                     <0, points[n_points-2].v, 0>
                     <cos_phi2*points[n_points-2].u, points[n_points-2].v, sin_phi2*points[n_points-2].u>
                     <cos_phi*points[n_points-2].u, points[n_points-2].v, sin_phi*points[n_points-2].u>
                   }
        #end              
        mor_do_face( <cos_phi*points[i].u, points[i].v, sin_phi*points[i].u>, <cos_phi*normals[i].u, normals[i].v, sin_phi*normals[i].u>,
                     <cos_phi*points[i+1].u, points[i+1].v, sin_phi*points[i+1].u>, <cos_phi*normals[i+1].u, normals[i+1].v, sin_phi*normals[i+1].u>,
                     <cos_phi2*points[i+1].u, points[i+1].v, sin_phi2*points[i+1].u>, <cos_phi2*normals[i+1].u, normals[i+1].v, sin_phi2*normals[i+1].u>,
                     <cos_phi2*points[i].u, points[i].v, sin_phi2*points[i].u>, <cos_phi2*normals[i].u, normals[i].v, sin_phi2*normals[i].u> )
        #local j=j+1;
      #end  
      #local i=i+1;
    #end  
  #end      
#end      

#macro mesh_of_revolution(points, angular_range, division, smoothing, closed)
  #ifndef (mor_do_face)
    #macro mor_do_face (corner_1, normal_1, corner_2, normal_2, corner_3, normal_3, corner_4, normal_4)
      #if (smoothing=0)
        triangle { corner_1, corner_2, corner_3 }
        triangle { corner_1, corner_3, corner_4 }
      #else
        smooth_triangle {corner_1, normal_1, corner_2, normal_2, corner_3, normal_3}   
        smooth_triangle {corner_1, normal_1, corner_3, normal_3, corner_4, normal_4}   
      #end  
    #end
  #end    
  mesh { mor_do_revolution(points, angular_range, division, smoothing, closed) }
#end      

#macro union_of_revolution(points, angular_range, division, smoothing, closed)
  #ifndef (mor_do_face)
    #macro mor_do_face (corner_1, normal_1, corner_2, normal_2, corner_3, normal_3, corner_4, normal_4)
      #if (smoothing=0)
        triangle { corner_1, corner_2, corner_3 }
        triangle { corner_1, corner_3, corner_4 }
      #else
        smooth_triangle {corner_1, normal_1, corner_2, normal_2, corner_3, normal_3}   
        smooth_triangle {corner_1, normal_1, corner_3, normal_3, corner_4, normal_4}   
      #end  
    #end
  #end    
  union { mor_do_revolution(points, angular_range, division, smoothing, closed) }
#end      